/* packet-ospf.c
 * Routines for OSPF packet disassembly
 * (c) Copyright Hannes R. Boehm <hannes@boehm.org>
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@zing.org>
 * Copyright 1998 Gerald Combs
 * 
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
 
#include "config.h"

#include <gtk/gtk.h>

#include <stdio.h>

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#include "ethereal.h"
#include "packet.h"
#include "packet-ospf.h"


void 
dissect_ospf(const u_char *pd, int offset, frame_data *fd, GtkTree *tree) {
    e_ospfhdr *ospfh;

    GtkWidget *ospf_tree = NULL, *ti; 
    GtkWidget *ospf_header_tree;
    char auth_data[9]="";
    char *packet_type;

    ospfh = (e_ospfhdr *) &pd[offset];

    switch(ospfh->packet_type) {
        case OSPF_HELLO:
	    packet_type="Hello Packet";
            break;
        case OSPF_DB_DESC:
	    packet_type="DB Descr.";
            break;
        case OSPF_LS_REQ:
	    packet_type="LS Request";
            break;
        case OSPF_LS_UPD:
	    packet_type="LS Update";
            break;
        case OSPF_LS_ACK:
	    packet_type="LS Acknowledge";
            break;
            default:
    }
    if (fd->win_info[0]) {
        strcpy(fd->win_info[3], "OSPF");
        sprintf(fd->win_info[4], "%s (%d)", packet_type ,ospfh->packet_type);
    }  

    if (tree) {
	ti = add_item_to_tree(GTK_WIDGET(tree), offset, ntohs(ospfh->length), "Open Shortest Path First"); 
	ospf_tree = gtk_tree_new(); 
	gtk_tree_item_set_subtree(GTK_TREE_ITEM(ti), ospf_tree);

	ti = add_item_to_tree(GTK_WIDGET(ospf_tree), offset, OSPF_HEADER_LENGTH, "OSPF Header"); 
	ospf_header_tree = gtk_tree_new();
	gtk_tree_item_set_subtree(GTK_TREE_ITEM(ti), ospf_header_tree);

        add_item_to_tree(ospf_header_tree, offset, 1, "OSPF Version: %d", ospfh->version);  
	add_item_to_tree(ospf_header_tree, offset + 1 , 1, "OSPF Packet Type: %d (%s)", 
	                                                   ospfh->packet_type, packet_type);
	add_item_to_tree(ospf_header_tree, offset + 2 , 2, "Packet Legth: %d", 
	                                                   ntohs(ospfh->length));
	add_item_to_tree(ospf_header_tree, offset + 4 , 4, "Source OSPF Router ID: %s", 
	                                                   ip_to_str((guint8 *) &(ospfh->routerid)));
	if (!(ospfh->area)) {
	   add_item_to_tree(ospf_header_tree, offset + 8 , 4, "Area ID: Backbone");
	} else {
	   add_item_to_tree(ospf_header_tree, offset + 8 , 4, "Area ID: %s", ip_to_str((guint8 *) &(ospfh->area)));
	}
	add_item_to_tree(ospf_header_tree, offset + 12 , 2, "Packet Checksum");
	switch( ntohs(ospfh->auth_type) ) {
	    case OSPF_AUTH_NONE:
	         add_item_to_tree(ospf_header_tree, offset + 14 , 2, "Auth Type: none");
	         add_item_to_tree(ospf_header_tree, offset + 16 , 8, "Auth Data (none)");
		 break;
	    case OSPF_AUTH_SIMPLE:
	         add_item_to_tree(ospf_header_tree, offset + 14 , 2, "Auth Type: simple");
                 strncpy(auth_data, &(ospfh->auth_data), 8);
	         add_item_to_tree(ospf_header_tree, offset + 16 , 8, "Auth Data: %s", auth_data);
		 break;
	    case OSPF_AUTH_CRYPT:
	         add_item_to_tree(ospf_header_tree, offset + 14 , 2, "Auth Type: crypt");
	         add_item_to_tree(ospf_header_tree, offset + 16 , 8, "Auth Data (crypt)");
		 break;
            default:
	         add_item_to_tree(ospf_header_tree, offset + 14 , 2, "Auth Type (unknown)");
	         add_item_to_tree(ospf_header_tree, offset + 16 , 8, "Auth Data (unknown)");
	}

    }

    /*  Skip over header */
    offset += OSPF_HEADER_LENGTH;
    switch(ospfh->packet_type){
	case OSPF_HELLO:
	    dissect_ospf_hello(pd, offset, fd, (GtkTree *) ospf_tree); 
	    break;
	case OSPF_DB_DESC:
	    dissect_ospf_db_desc(pd, offset, fd, (GtkTree *) ospf_tree);   
	    break;
	case OSPF_LS_REQ:
	    dissect_ospf_ls_req(pd, offset, fd, (GtkTree *) ospf_tree);   
	    break;
	case OSPF_LS_UPD:
	    dissect_ospf_ls_upd(pd, offset, fd, (GtkTree *) ospf_tree);
	    break;
	case OSPF_LS_ACK:
	default:
            dissect_raw(pd, offset, fd, tree); 
    }
}

void
dissect_ospf_hello(const u_char *pd, int offset, frame_data *fd, GtkTree *tree) {
    e_ospf_hello *ospfhello;
    guint32 *ospfneighbor;
    char options[20]="";
    int options_offset;

    GtkWidget *ospf_hello_tree, *ti; 

    ospfhello = (e_ospf_hello *) &pd[offset];

    if (tree) {
	ti = add_item_to_tree(GTK_WIDGET(tree), offset, (fd->cap_len - offset) , "OSPF Hello Packet"); 
	ospf_hello_tree = gtk_tree_new(); 
	gtk_tree_item_set_subtree(GTK_TREE_ITEM(ti), ospf_hello_tree);


	add_item_to_tree(ospf_hello_tree, offset , 4, "Network Mask: %s",  ip_to_str((guint8 *) &ospfhello->network_mask));
	add_item_to_tree(ospf_hello_tree, offset + 4, 2, "Hello Intervall: %d seconds",  ntohs(ospfhello->hellointervall));


	/* ATTENTION !!! no check for length of options string */
	options_offset=0;
	if(( ospfhello->options & OSPF_OPTIONS_E ) == OSPF_OPTIONS_E){
	    strcpy( (char *)(options + options_offset), "E");
	    options_offset+=1;
	}
	if(( ospfhello->options & OSPF_OPTIONS_MC ) == OSPF_OPTIONS_MC){
	    strcpy((char *) (options + options_offset), "/MC");
	    options_offset+=3;
	}
	if(( ospfhello->options & OSPF_OPTIONS_NP ) == OSPF_OPTIONS_NP){
	    strcpy((char *) (options + options_offset), "/NP");
	    options_offset+=3;
	}
	if(( ospfhello->options & OSPF_OPTIONS_EA ) == OSPF_OPTIONS_EA){
	    strcpy((char *) (options + options_offset) , "/EA");
	    options_offset+=3;
	}
	if(( ospfhello->options & OSPF_OPTIONS_DC ) == OSPF_OPTIONS_DC){
	    strcpy((char *) (options + options_offset) , "/DC");
	    options_offset+=3;
	}

	add_item_to_tree(ospf_hello_tree, offset + 6, 1, "Options: %d (%s)",  ospfhello->options, options);
	add_item_to_tree(ospf_hello_tree, offset + 7, 1, "Router Priority: %d",  ospfhello->priority);
	add_item_to_tree(ospf_hello_tree, offset + 8, 4, "RouterDeadIntervall: %d seconds",  ntohl(ospfhello->dead_interval));
	add_item_to_tree(ospf_hello_tree, offset + 12, 4, "Designated Router: %s",  ip_to_str((guint8 *) &ospfhello->drouter));
	add_item_to_tree(ospf_hello_tree, offset + 16, 4, "Backup Designated Router: %s",  ip_to_str((guint8 *) &ospfhello->bdrouter));

	offset+=20;
	while((fd->cap_len - offset) >= 4){
	    ospfneighbor=(guint32 *) &pd[offset];
	    add_item_to_tree(ospf_hello_tree, offset, 4, "Active Neighbor: %s",  ip_to_str((guint8 *) ospfneighbor));
	    offset+=4;
	}
    }
}

void
dissect_ospf_db_desc(const u_char *pd, int offset, frame_data *fd, GtkTree *tree) {
    e_ospf_dbd *ospf_dbd;
    char options[20]="";
    int options_offset;
    char flags[20]="";
    int flags_offset;

    GtkWidget *ospf_db_desc_tree, *ti; 

    ospf_dbd = (e_ospf_dbd *) &pd[offset]; 

    if (tree) {
	ti = add_item_to_tree(GTK_WIDGET(tree), offset, (fd->cap_len - offset) , "OSPF DB Description"); 
	ospf_db_desc_tree = gtk_tree_new(); 
	gtk_tree_item_set_subtree(GTK_TREE_ITEM(ti), ospf_db_desc_tree);

	add_item_to_tree(ospf_db_desc_tree, offset, 2, "Interface MTU: %d", ntohs(ospf_dbd->interface_mtu) );


	options_offset=0;
	if(( ospf_dbd->options & OSPF_OPTIONS_E ) == OSPF_OPTIONS_E){
	    strcpy( (char *)(options + options_offset), "_E_");
	    options_offset+=1;
	}
	if(( ospf_dbd->options & OSPF_OPTIONS_MC ) == OSPF_OPTIONS_MC){
	    strcpy((char *) (options + options_offset), "_MC_");
	    options_offset+=3;
	}
	if(( ospf_dbd->options & OSPF_OPTIONS_NP ) == OSPF_OPTIONS_NP){
	    strcpy((char *) (options + options_offset), "_NP_");
	    options_offset+=3;
	}
	if(( ospf_dbd->options & OSPF_OPTIONS_EA ) == OSPF_OPTIONS_EA){
	    strcpy((char *) (options + options_offset) , "_EA_");
	    options_offset+=3;
	}
	if(( ospf_dbd->options & OSPF_OPTIONS_DC ) == OSPF_OPTIONS_DC){
	    strcpy((char *) (options + options_offset) , "_DC_");
	    options_offset+=3;
	}

	add_item_to_tree(ospf_db_desc_tree, offset + 2 , 1, "Options: %d (%s)", ospf_dbd->options, options );


	flags_offset=0;
	if(( ospf_dbd->flags & OSPF_DBD_FLAG_MS ) == OSPF_DBD_FLAG_MS){
	    strcpy( (char *)(flags + flags_offset), "_I_");
	    flags_offset+=1;
	}
	if(( ospf_dbd->flags & OSPF_DBD_FLAG_M ) == OSPF_DBD_FLAG_M){
	    strcpy((char *) (flags + flags_offset), "_M_");
	    flags_offset+=3;
	}
	if(( ospf_dbd->flags & OSPF_DBD_FLAG_I ) == OSPF_DBD_FLAG_I){
	    strcpy((char *) (flags + flags_offset), "_I_");
	    flags_offset+=3;
	}

	add_item_to_tree(ospf_db_desc_tree, offset + 3 , 1, "Flags: %d (%s)", ospf_dbd->flags, flags );
	add_item_to_tree(ospf_db_desc_tree, offset + 4 , 4, "DD Sequence: %d", ntohl(ospf_dbd->dd_sequence) );
    }
    /* LS Headers will be processed here */
}

void
dissect_ospf_ls_req(const u_char *pd, int offset, frame_data *fd, GtkTree *tree) {
    e_ospf_ls_req *ospf_lsr;

    GtkWidget *ospf_lsr_tree, *ti; 


    /* zero or more LS requests may follow */
    /* each LSA-Header consists of 4bytes LS-type, 4bytes LS-id, 4bytes AdvRouter = 12 bytes*/
    /* we place every request for a LSA in a single subtree */
    if (tree) {
	while( ( fd->cap_len - offset) >= OSPF_LS_REQ_LENGTH ){
             ospf_lsr = (e_ospf_ls_req *) &pd[offset]; 
	     ti = add_item_to_tree(GTK_WIDGET(tree), offset, OSPF_LS_REQ_LENGTH, "Link State Request"); 
	     ospf_lsr_tree = gtk_tree_new(); 
	     gtk_tree_item_set_subtree(GTK_TREE_ITEM(ti), ospf_lsr_tree);

	     switch( ntohl( ospf_lsr->ls_type ) ){
		 case OSPF_LSTYPE_ROUTER:
	             add_item_to_tree(ospf_lsr_tree, offset, 4, "LS Type: Router-LSA (%d)", 
	                               ntohl( ospf_lsr->ls_type ) );
		     break;
		 case OSPF_LSTYPE_NETWORK:
	             add_item_to_tree(ospf_lsr_tree, offset, 4, "LS Type: Network-LSA (%d)", 
	                               ntohl( ospf_lsr->ls_type ) );
		     break;
		 case OSPF_LSTYPE_SUMMERY:
	             add_item_to_tree(ospf_lsr_tree, offset, 4, "LS Type: Summary-LSA (IP network) (%d)", 
	                               ntohl( ospf_lsr->ls_type ) );
		     break;
		 case OSPF_LSTYPE_ASBR:
	             add_item_to_tree(ospf_lsr_tree, offset, 4, "LS Type: Summary-LSA (ASBR) (%d)", 
	                               ntohl( ospf_lsr->ls_type ) );
		     break;
		 case OSPF_LSTYPE_ASEXT:
	             add_item_to_tree(ospf_lsr_tree, offset, 4, "LS Type: AS-External-LSA (ASBR) (%d)", 
	                               ntohl( ospf_lsr->ls_type ) );
		     break;
		 default:
	             add_item_to_tree(ospf_lsr_tree, offset, 4, "LS Type: %d (unknown)", 
	                               ntohl( ospf_lsr->ls_type ) );
	     }

             add_item_to_tree(ospf_lsr_tree, offset + 4, 4, "Link State ID : %s", 
	                                 ip_to_str((guint8 *) &(ospf_lsr->ls_id)));
             add_item_to_tree(ospf_lsr_tree, offset + 4, 4, "Advertising Router : %s", 
	                                 ip_to_str((guint8 *) &(ospf_lsr->adv_router)));

	     offset+=12;
	}
    }
}
void
dissect_ospf_ls_upd(const u_char *pd, int offset, frame_data *fd, GtkTree *tree) {
}
